'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.nah = undefined;
exports.defineAtom = defineAtom;
exports.defineFormula = defineFormula;
exports.defineValueAtom = defineValueAtom;
exports.atomify = atomify;
exports.combine = combine;
exports.sink = sink;
exports.atomConnector = atomConnector;
exports.atomPropConnector = atomPropConnector;
exports.actionSwitch = actionSwitch;
exports.bindActionCreator = bindActionCreator;
exports.lazify = lazify;
exports.link = link;
exports.mapObject = mapObject;
exports.merge = merge;

var _zine = require('zine');

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

// Constants

var nah = exports.nah = {};

// Atom defining functions

function defineAtom(reducer, initialState) {
  var state = initialState;

  function atom() {
    return state;
  }

  atom.update = function (value) {
    if (value !== nah) {
      var newState = reducer(state, value);
      if (newState != nah) {
        state = newState;
        (0, _zine.publish)(atom, state);
      }
    }
    return state;
  };

  return atom;
};

function defineFormula(formula) {
  var state; // set below
  var values;

  function boundFormula(values) {
    return state;
  }

  for (var _len = arguments.length, atoms = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    atoms[_key - 1] = arguments[_key];
  }

  values = atoms.map(function (atom, index) {
    (0, _zine.subscribe)(atom, function (newValue) {
      if (newValue != nah) {
        values[index] = newValue;
        var newState = formula.apply(undefined, _toConsumableArray(values));
        if (newState != nah) {
          state = newState;
          (0, _zine.publish)(boundFormula, state);
        }
      }
    });
    return atom();
  });

  state = formula.apply(undefined, _toConsumableArray(values));
  return boundFormula;
}

function defineValueAtom(initialState) {
  var state = initialState;

  function atom() {
    return state;
  }

  atom.update = function (value) {
    if (value !== nah) {
      state = value;
      (0, _zine.publish)(atom, state);
    }
    return state;
  };

  return atom;
};

// Atom definition helpers

function atomify(fn, initialState) {
  return defineAtom(function (unused, value) {
    return fn(value);
  });
}

function combine(atomMap) {
  var state = {};

  function combinator() {
    return state;
  }

  var _loop = function _loop(key) {
    atom = atomMap[key];

    (0, _zine.subscribe)(atom, function (newValue) {
      if (newValue != nah) {
        state = merge(state, _defineProperty({}, key, newValue));
        (0, _zine.publish)(combinator, state);
      }
    });
    state[key] = atom();
  };

  for (var key in atomMap) {
    var atom;

    _loop(key);
  }

  return combinator;
}

function sink() {
  for (var _len2 = arguments.length, atoms = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    atoms[_key2] = arguments[_key2];
  }

  return defineFormula.apply(undefined, [merge].concat(atoms));
}

// Connectors

function call(sub) {
  // not exported
  return sub();
}

function atomConnector(subscriptionSpec) {
  return (0, _zine.connector)(subscriptionSpec, call);
}

function atomPropConnector(prop) {
  return (0, _zine.propConnector)(prop, call);
}

// Utility functions

function actionSwitch(actionMap) {
  // TODO: verify that all keys are functions?
  return function (state, action) {
    var caseFn = actionMap[action.type];
    return caseFn ? caseFn(state, action) : nah;
  };
}

function bindActionCreator(dispatch, fn) {
  if (fn) {
    return function () {
      return dispatch(fn.apply(undefined, arguments));
    };
  } else {
    return function (func) {
      return bindActionCreator(dispatch, func);
    };
  }
}

function lazify(fn) {
  return function (state, value) {
    var newState = fn(state, value);
    return newState === state ? nah : newState;
  };
}

function link() {
  for (var _len3 = arguments.length, atoms = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
    atoms[_key3] = arguments[_key3];
  }

  return atoms.reduce(function (a, b) {
    (0, _zine.subscribe)(a, b.update);
    return b;
  });
}

function mapObject(func, obj) {
  var newObj = {};
  for (var key in obj) {
    newObj[key] = func(obj[key], key);
  }
  return newObj;
}

function merge() {
  for (var _len4 = arguments.length, objects = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
    objects[_key4] = arguments[_key4];
  }

  return Object.assign.apply(Object, [{}].concat(objects));
}

exports.default = { nah: nah, defineAtom: defineAtom, defineFormula: defineFormula, defineValueAtom: defineValueAtom, atomify: atomify, combine: combine, sink: sink, atomConnector: atomConnector, atomPropConnector: atomPropConnector, actionSwitch: actionSwitch, bindActionCreator: bindActionCreator, lazify: lazify, link: link, mapObject: mapObject, merge: merge };
